""" world.py
 Copyright (C) 1998 Aloril
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""

import whrandom
from event import *
from thing import thing
from mind import mind
from copy import deepcopy
import dictlist, const, types, geometry,wtime

#whole world
class world:
    "whole world"
    def __init__(self):
        self.things={}
        self.minds=[]
        self.time=wtime.world_time()
        self.events=[]
        self.event_dict={}
        self.id_dict={}
    def add_thing(self,thing):
        dictlist.add_value(self.things,thing.name,thing)
        id="_"+thing.name
        i=1
        while self.id_dict.has_key(id):
            id="_"+thing.name+`i`
            i=i+1
        thing.id=id
        self.id_dict[id]=thing
        self.events.append(event("make",what=thing))
    def add_mind(self,mind):
        self.minds.append(mind)
        self.add_thing(mind)
    def remove_thing(self, thing):
        dictlist.remove_value(self.things, thing)
        if thing in self.minds: self.minds.remove(thing)
        del self.id_dict[thing.id]
    def get_thing_list(self):
        res=[]
        for tl in self.things.values():
            res=res+tl
        return res
    def __str__(self):
        s=str(self.time)+"\n"
        for m in self.minds: s=s+str(m)
        return s
    def __repr__(self): return `"world"`
    def destroy(self):
        for tl in self.things.values():
            for t in tl:
                t.destroy()
##     def look_all(self):
##         res=[]
##         for t in self.get_thing_list():
##             res.append(event('look',what=t.id,loc=t.get_xyz()))
##         return res
    #event handling: 
    #m has current mind, e is current event(=output) from that mind
    #may update self.events
    def move_event(self,m,e):
        e2=event("move",event_source=m,what=e.what)
        w=e.what
        if not self.id_dict.has_key(w.id):
            self.events.append(event("illegal",what=e,what_desc="e.what.id"))
            return
        #source
        source_xyz=w.get_xyz()
        if type(source_xyz)!=types.TupleType:
            self.events.append(event("illegal",what=e,what_desc="source_xyz"))
            return
        source_xyz=tuple(map(float,source_xyz))
        #target
        try:
            target_xyz=e.loc.get_xyz()
        except AttributeError:
            target_xyz=e.loc
            if type(target_xyz)!=types.TupleType:
                self.events.append(event("illegal",what=e,what_desc="target_xyz"))
                return
        target_xyz=tuple(map(float,target_xyz))
        #decide how far
        if geometry.distance(source_xyz,target_xyz)>100.0:
            target=geometry.distance_from_begin(source_xyz,target_xyz,100.0)
            w.xyz=target
            w.place=None
        else:
            target=e.loc
            #if its part of inventory, use its xyz
            if hasattr(target,"xyz"):
                if target.xyz:
                    w.xyz=None
                w.place=target
            else:
                w.xyz=target_xyz
                w.place=None
        e2.loc=target
        self.events.append(e2)
    def cut_event(self,m,e):
        t=e.what #thing
        t.status=t.status-0.1
        if t.status<=const.fzero:
            self.events.append(event("destroy", event_source=m, what=t))
            self.remove_thing(t)
    def make_event(self,m,e):
        if types.StringType==type(e.what):
            if e.loc:
                t=thing(e.what,e.what_desc,e.what_desc,0.1,xyz=e.loc)
            else:
                t=thing(e.what,e.what_desc,m,0.1)
            self.add_thing(t)
        else:
            t=e.what
            t.status=t.status+0.1
        self.events.append(event("make", event_source=m, what=t))
    def say_event(self,m,e):
        self.events.append(event("say", event_source=m, what=e.what))
    def extinguish_event(self,m,e):
        f=e.loc
        f.status=f.status-0.1
        if f.status<=const.fzero:
            self.events.append(event("destroy",event_source=self, what=f))
            self.remove_thing(f)
##     def look_event(self,m,e):
##         try:
##             t=self.id_dict[e.what]
##             self.events.append(event("look",event_source=m,
##                                      what=ething(t.id,name=t.name,type=t.get_type())))
##         except KeyError:
##             pass
    def time_event(self,m,e):
        self.events.append(event("time",what=str(self.time)))
    def undefined_event(self,m,e):
        #print "Undefined event:", e.command
        pass
        #else:
        #    e.event_source=m
        #    self.events.append(e)
    def process_mind_output(self,m,e):
        """m has current mind, e is current event(=output) from that mind
           may update self.event"""
        get_event_func(self,e)(m,e)
        return self.events
    def run_step(self):
        #age all things
        self.time.step()
        for tl in self.things.values():
            for t in tl: t.age=t.age+self.time.stepseconds
        #add sight events about houses
        #CHEAT!: not effective and doesn't take into account range or obstacles
        for h in self.things.get('house',[]):
            self.events.append(event("sight",what=h))
        #let minds think
        output=[]
        for m in self.minds: 
            output.append([m,m.run_step(self.time,self.events)])
        self.events=[]
        for (m,mo) in output:
            for e in mo:
                #m has current mind, e is current event(=output) from that mind
                self.process_mind_output(m,e)
        #random events
        #birth (CHEAT!: should use some change/time/count of objects)
        if whrandom.randint(0,300)==0:
            #currently no genders
            ml=filter(lambda m:m.age>=wtime.years(18),self.minds)
            if ml:
                m=whrandom.choice(ml)
                #it gives birth
                #CHEAT!: name should be given by parents!
                #do it when genrally implement talk/rumors
                i=1
                while 1:
                    name=m.name+`i`
                    if not self.things.has_key(name): break
                    i=i+1
                mnew=mind(name,desc=m.__doc__,place=m.place,
                          tasks=m.tasks, xyz=m.xyz,
                          personal_knowledge=deepcopy(m.personal_knowledge),
                          local_knowledge=m.local_knowledge,
                          common_knowledge=m.common_knowledge,
                          common_goals=m.common_goals)
                self.add_mind(mnew)
                self.events.append(event("make", desc="birth",
                                         event_source=self, what=mnew))
        #death (CHEAT!: should use some change/time/count of objects)
        #should take into account age too
        if whrandom.randint(0,400)==0:
            if self.minds:
                m=whrandom.choice(self.minds)
                e=event("destroy", desc="death", event_source=self, what=m)
                self.events.append(e)
                #CHEAT! : leak: add things back to world
                m.run_step(self.time,[e])
                self.remove_thing(m)
        #fire (CHEAT!: should use some change/time/count of objects)
        if whrandom.randint(0,300)==0 and not self.things.get('fire'):
            hl=self.things.get('house')
            if hl:
                h=whrandom.choice(hl)
                #set fire to house
                f=thing('fire','semilive thing',h,0.0)
                self.add_thing(f)
                self.events.append(event("make",event_source=self, what=f))
        #run things run by world
        #'run' fire
        for f in self.things.get('fire',[]):
            t=f.place
            f.status=f.status+0.15
            if f.status>1.0:
                self.events.append(event("destroy", event_source=self, what=t))
                self.remove_thing(t)
                self.remove_thing(f)
            else:
                self.events.append(event("fire", event_source=self, what=f, loc=t))
        return self.events
